import platform

from smartcard.pcsc.PCSCExceptions import *
from smartcard.scard import *


def test_list_readers_exception():
    exc = ListReadersException(0)
    assert exc.hresult == 0
    text = str(exc)
    if platform.system() == "Windows":
        expected = (
            "Failed to list readers: "
            "The operation completed successfully.  (0x00000000)"
        )
    else:
        expected = "Failed to list readers: Command successful. (0x00000000)"
    assert text == expected

    exc = ListReadersException(0x42)
    assert exc.hresult == 0x42
    text = str(exc)
    if platform.system() == "Windows":
        expected = (
            "Failed to list readers: "
            "The network resource type is not correct.  (0x00000042)"
        )
    else:
        expected = "Failed to list readers: Unknown error: 0x00000042 (0x00000042)"
    macos_bug_expected = expected.replace("Unknown", "Unkown")
    assert text in (expected, macos_bug_expected)

    exc = ListReadersException(SCARD_S_SUCCESS)
    assert exc.hresult == 0

    exc = ListReadersException(SCARD_E_NO_SERVICE)
    assert exc.hresult == SCARD_E_NO_SERVICE
    text = str(exc)
    if platform.system() == "Windows":
        expected = (
            "Failed to list readers: "
            "The Smart Card Resource Manager is not running.  (0x8010001D)"
        )
    else:
        expected = "Failed to list readers: Service not available. (0x8010001D)"
    assert text == expected


def test_establish_context_exception():
    exc = EstablishContextException(SCARD_E_NO_SERVICE)
    assert exc.hresult == SCARD_E_NO_SERVICE
    text = str(exc)
    if platform.system() == "Windows":
        expected = (
            "Failed to establish context: "
            "The Smart Card Resource Manager is not running.  (0x8010001D)"
        )
    else:
        expected = "Failed to establish context: Service not available. (0x8010001D)"
    assert text == expected


def test_introduce_reader_exception():
    exc = IntroduceReaderException(SCARD_E_DUPLICATE_READER, "foobar")
    assert exc.hresult == SCARD_E_DUPLICATE_READER
    text = str(exc)
    if platform.system() == "Windows":
        expected = (
            "Failed to introduce a new reader: foobar: "
            "The reader driver did not produce a unique reader name.  (0x8010001B)"
        )
    else:
        expected = (
            "Failed to introduce a new reader: foobar: "
            "Reader already exists. (0x8010001B)"
        )
    assert text == expected


def test_remove_reader_from_group_exception():
    exc = RemoveReaderFromGroupException(
        SCARD_E_INVALID_HANDLE, "readername", "readergroup"
    )
    assert exc.hresult == SCARD_E_INVALID_HANDLE
    text = str(exc)
    if platform.system() == "Windows":
        expected = (
            "Failed to remove reader: readername from group: readergroup: "
            "The supplied handle was invalid.  (0x80100003)"
        )
    else:
        expected = (
            "Failed to remove reader: readername from group: readergroup: "
            "Invalid handle. (0x80100003)"
        )
    assert text == expected


def test_add_reader_to_group_exception():
    exc = AddReaderToGroupException(SCARD_E_INVALID_HANDLE, "reader", "group")
    assert exc.hresult == SCARD_E_INVALID_HANDLE
    text = str(exc)
    if platform.system() == "Windows":
        expected = (
            "Failed to add reader: reader to group: group: "
            "The supplied handle was invalid.  (0x80100003)"
        )
    else:
        expected = (
            "Failed to add reader: reader to group: group: "
            "Invalid handle. (0x80100003)"
        )
    assert text == expected


def test_release_context_exception():
    exc = ReleaseContextException(SCARD_E_INVALID_HANDLE)
    assert exc.hresult == SCARD_E_INVALID_HANDLE
    text = str(exc)
    if platform.system() == "Windows":
        expected = (
            "Failed to release context: The supplied handle was invalid.  (0x80100003)"
        )
    else:
        expected = "Failed to release context: Invalid handle. (0x80100003)"
    assert text == expected


def test_base_scard_exception():
    exc = BaseSCardException(SCARD_E_UNKNOWN_READER)
    assert exc.hresult == SCARD_E_UNKNOWN_READER
    text = str(exc)
    if platform.system() == "Windows":
        expected = (
            "scard exception: "
            "The specified reader name is not recognized.  (0x80100009)"
        )
    else:
        expected = "scard exception: Unknown reader specified. (0x80100009)"
    assert text == expected

    exc = BaseSCardException(hresult=-1)
    assert exc.hresult == -1
    text = str(exc)
    expected = "scard exception"
    assert text == expected

    exc = BaseSCardException(hresult=-1, message="foo bar")
    assert exc.hresult == -1
    text = str(exc)
    expected = "foo bar"
    assert text == expected

    exc = BaseSCardException(message="foo", hresult=SCARD_E_NOT_TRANSACTED)
    assert exc.hresult == SCARD_E_NOT_TRANSACTED
    text = str(exc)
    if platform.system() == "Windows":
        expected = "An attempt was made to end a non-existent transaction. "
    else:
        expected = "Transaction failed."
    expected = "foo: " + expected + " (0x80100016)"
    assert text == expected



"""Smart card reader monitoring classes.

ReaderObserver is a base class for objects that are to be notified
upon smartcard reader insertion/removal.

ReaderMonitor is a singleton object notifying registered ReaderObservers
upon reader insertion/removal.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import traceback
from threading import Event, Thread
from time import sleep

import smartcard.System
from smartcard.Observer import Observable, Observer
from smartcard.Synchronization import *


# ReaderObserver interface
class ReaderObserver(Observer):
    """
    ReaderObserver is a base abstract class for objects that are to be notified
    upon smartcard reader insertion/removal.
    """

    def __init__(self):
        pass

    def update(self, observable, handlers):
        """Called upon reader insertion/removal.

        @param observable:
        @param handlers:
          - addedreaders: list of added readers causing notification
          - removedreaders: list of removed readers causing notification
        """
        pass


class ReaderMonitor(Observable):
    """Class that monitors reader insertion/removal.
    and notify observers

    note: a reader monitoring thread will be running
    as long as the reader monitor has observers, or ReaderMonitor.stop()
    is called.

    It implements the shared state design pattern, where objects
    of the same type all share the same state, in our case essentially
    the ReaderMonitoring Thread. Thanks to Frank Aune for implementing
    the shared state pattern logics.
    """

    __shared_state = {}

    def __init__(
        self, startOnDemand=True, readerProc=smartcard.System.readers, period=1
    ):
        self.__dict__ = self.__shared_state
        Observable.__init__(self)
        self.startOnDemand = startOnDemand
        self.readerProc = readerProc
        self.period = period
        if self.startOnDemand:
            self.rmthread = None
        else:
            self.rmthread = ReaderMonitoringThread(self, self.readerProc, self.period)
            self.rmthread.start()

    def addObserver(self, observer):
        """Add an observer."""
        Observable.addObserver(self, observer)

        # If self.startOnDemand is True, the reader monitoring
        # thread only runs when there are observers.
        if self.startOnDemand:
            if 0 < self.countObservers():
                if not self.rmthread:
                    self.rmthread = ReaderMonitoringThread(
                        self, self.readerProc, self.period
                    )

                    # start reader monitoring thread in another thread to
                    # avoid a deadlock; addObserver and notifyObservers called
                    # in the ReaderMonitoringThread run() method are
                    # synchronized
                    import _thread

                    _thread.start_new_thread(self.rmthread.start, ())
        else:
            observer.update(self, (self.rmthread.readers, []))

    def deleteObserver(self, observer):
        """Remove an observer."""
        Observable.deleteObserver(self, observer)
        # If self.startOnDemand is True, the reader monitoring
        # thread is stopped when there are no more observers.
        if self.startOnDemand:
            if 0 == self.countObservers():
                self.rmthread.stop()
                del self.rmthread
                self.rmthread = None

    def __str__(self):
        return self.__class__.__name__


synchronize(
    ReaderMonitor,
    "addObserver deleteObserver deleteObservers "
    + "setChanged clearChanged hasChanged "
    + "countObservers",
)


class ReaderMonitoringThread(Thread):
    """Reader insertion thread.
    This thread polls for pcsc reader insertion, since no
    reader insertion event is available in pcsc.
    """

    __shared_state = {}

    def __init__(self, observable, readerProc, period):
        self.__dict__ = self.__shared_state
        Thread.__init__(self)
        self.observable = observable
        self.stopEvent = Event()
        self.stopEvent.clear()
        self.readers = []
        self.daemon = True
        self.name = "smartcard.ReaderMonitoringThread"
        self.readerProc = readerProc
        self.period = period

    def run(self):
        """Runs until stopEvent is notified, and notify
        observers of all reader insertion/removal.
        """
        while not self.stopEvent.is_set():
            try:
                # no need to monitor if no observers
                if 0 < self.observable.countObservers():
                    currentReaders = self.readerProc()
                    addedReaders = []
                    removedReaders = []

                    if currentReaders != self.readers:
                        for reader in currentReaders:
                            if reader not in self.readers:
                                addedReaders.append(reader)
                        for reader in self.readers:
                            if reader not in currentReaders:
                                removedReaders.append(reader)

                        if addedReaders or removedReaders:
                            # Notify observers
                            self.readers = []
                            for r in currentReaders:
                                self.readers.append(r)
                            self.observable.setChanged()
                            self.observable.notifyObservers(
                                (addedReaders, removedReaders)
                            )

                # wait every second on stopEvent
                self.stopEvent.wait(self.period)

            except Exception:
                # FIXME Tighten the exceptions caught by this block
                traceback.print_exc()
                # Most likely raised during interpreter shutdown due
                # to unclean exit which failed to remove all observers.
                # To solve this, we set the stop event and pass the
                # exception to let the thread finish gracefully.
                self.stopEvent.set()

    def stop(self):
        self.stopEvent.set()
        self.join()


if __name__ == "__main__":
    print("insert or remove readers in the next 20 seconds")

    # a simple reader observer that prints added/removed readers
    class printobserver(ReaderObserver):

        def __init__(self, obsindex):
            self.obsindex = obsindex

        def update(self, observable, handlers):
            addedreaders, removedreaders = handlers
            print("%d - added:   " % self.obsindex, addedreaders)
            print("%d - removed: " % self.obsindex, removedreaders)

    class testthread(Thread):

        def __init__(self, obsindex):
            Thread.__init__(self)
            self.readermonitor = ReaderMonitor()
            self.obsindex = obsindex
            self.observer = None

        def run(self):
            # create and register observer
            self.observer = printobserver(self.obsindex)
            self.readermonitor.addObserver(self.observer)
            sleep(20)
            self.readermonitor.deleteObserver(self.observer)

    t1 = testthread(1)
    t2 = testthread(2)
    t1.start()
    t2.start()
    t1.join()
    t2.join()



"""Sample CardConnectionDecorator that provides exclusive transmit()

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import smartcard.pcsc
from smartcard.CardConnectionDecorator import CardConnectionDecorator
from smartcard.Exceptions import CardConnectionException
from smartcard.scard import (
    SCARD_LEAVE_CARD,
    SCARD_S_SUCCESS,
    SCardBeginTransaction,
    SCardEndTransaction,
    SCardGetErrorMessage,
)


class ExclusiveTransmitCardConnection(CardConnectionDecorator):
    """This decorator uses
    L{SCardBeginTransaction}/L{SCardEndTransaction} to preserve other
    processes of threads to access the card during transmit()."""

    def __init__(self, cardconnection):
        CardConnectionDecorator.__init__(self, cardconnection)

    def lock(self):
        """Lock card with L{SCardBeginTransaction}."""

        component = self.component
        while True:
            if isinstance(
                component, smartcard.pcsc.PCSCCardConnection.PCSCCardConnection
            ):
                hresult = SCardBeginTransaction(component.hcard)
                if SCARD_S_SUCCESS != hresult:
                    raise CardConnectionException(
                        "Failed to lock with SCardBeginTransaction: "
                        + SCardGetErrorMessage(hresult)
                    )
                else:
                    # print('locked')
                    pass
                break
            if hasattr(component, "component"):
                component = component.component
            else:
                break

    def unlock(self):
        """Unlock card with L{SCardEndTransaction}."""
        component = self.component
        while True:
            if isinstance(
                component, smartcard.pcsc.PCSCCardConnection.PCSCCardConnection
            ):
                hresult = SCardEndTransaction(component.hcard, SCARD_LEAVE_CARD)
                if SCARD_S_SUCCESS != hresult:
                    raise CardConnectionException(
                        "Failed to unlock with SCardEndTransaction: "
                        + SCardGetErrorMessage(hresult)
                    )
                else:
                    # print('unlocked')
                    pass
                break
            if hasattr(component, "component"):
                component = component.component
            else:
                break

    def transmit(self, command, protocol=None):
        """Gain exclusive access to card during APDU transmission for if this
        decorator decorates a PCSCCardConnection."""
        data, sw1, sw2 = CardConnectionDecorator.transmit(self, command, protocol)
        return data, sw1, sw2



#! /usr/bin/env python3
"""Unit tests for smartcard.readers.ReaderGroups

This test case can be executed individually, or with all other test cases
thru testsuite_framework.py.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import unittest

from smartcard.scard import resourceManager
from smartcard.System import readergroups

if "winscard" == resourceManager:

    class testcase_readergroups(unittest.TestCase):
        """Test smartcard framework readersgroups."""

        pinpadgroup = "Pinpad$Readers"
        biogroup = "Biometric$Readers"

        def testcase_readergroup_add(self):
            """tests groups=groups+[newgroups]"""

            # take a snapshot of current groups
            groupssnapshot = list(readergroups())
            groups = readergroups()

            # add pinpad group
            groups = groups + [self.pinpadgroup]
            self.assertEqual(groups, groupssnapshot + [self.pinpadgroup])

            # add pinpad a second time and biometric once
            groups = groups + [self.biogroup, self.pinpadgroup]
            self.assertEqual(groups, groupssnapshot + [self.pinpadgroup, self.biogroup])

            # clean-up
            groups.remove(self.biogroup)
            groups.remove(self.pinpadgroup)

        def testcase_readergroup_iadd(self):
            """test groups+=[newgroups]"""

            # take a snapshot of current groups
            groupssnapshot = list(readergroups())
            groups = readergroups()

            # add pinpad group
            groups += [self.pinpadgroup]
            self.assertEqual(groups, groupssnapshot + [self.pinpadgroup])

            # add pinpad a second time and biometric once
            groups += [self.biogroup, self.pinpadgroup]
            self.assertEqual(groups, groupssnapshot + [self.pinpadgroup, self.biogroup])

            # clean-up
            groups.remove(self.biogroup)
            groups.remove(self.pinpadgroup)

        def testcase_readergroup_radd(self):
            """test groups=[newgroups]+groups"""

            # take a snapshot of current groups
            groupssnapshot = list(readergroups())
            groups = readergroups()

            # add pinpad group
            zgroups = [self.pinpadgroup] + groups
            self.assertEqual(groups, groupssnapshot)
            self.assertEqual(zgroups, groupssnapshot + [self.pinpadgroup])
            self.assertTrue(isinstance(zgroups, list))
            self.assertTrue(isinstance(groups, type(readergroups())))

            # add pinpad a tiwce and biometric once
            zgroups = [self.pinpadgroup, self.biogroup, self.pinpadgroup] + groups
            self.assertEqual(groups, groupssnapshot)
            self.assertEqual(
                zgroups, groupssnapshot + [self.pinpadgroup, self.biogroup]
            )
            self.assertTrue(isinstance(zgroups, list))
            self.assertTrue(isinstance(groups, type(readergroups())))

        def testcase_readergroup_append(self):
            """test groups.append(newgroups)"""

            # take a snapshot of current groups
            groupssnapshot = list(readergroups())
            groups = readergroups()

            # add pinpad group
            groups.append(self.pinpadgroup)
            self.assertEqual(groups, groupssnapshot + [self.pinpadgroup])

            # add pinpad a second time
            groups.append(self.pinpadgroup)
            self.assertEqual(groups, groupssnapshot + [self.pinpadgroup])

            # add biometric once
            groups.append(self.biogroup)
            self.assertEqual(groups, groupssnapshot + [self.pinpadgroup, self.biogroup])

            # clean-up
            groups.remove(self.biogroup)
            groups.remove(self.pinpadgroup)

        def testcase_readergroup_insert(self):
            """test groups.insert(i,newgroups)"""

            # take a snapshot of current groups
            groupssnapshot = list(readergroups())
            groups = readergroups()

            # add pinpad group
            groups.insert(0, self.pinpadgroup)
            self.assertEqual(groups, groupssnapshot + [self.pinpadgroup])

            # add pinpad a second time
            groups.insert(1, self.pinpadgroup)
            self.assertEqual(groups, groupssnapshot + [self.pinpadgroup])

            # add biometric once
            groups.insert(1, self.biogroup)
            self.assertEqual(groups, groupssnapshot + [self.pinpadgroup, self.biogroup])

            # clean-up
            groups.remove(self.biogroup)
            groups.remove(self.pinpadgroup)


if __name__ == "__main__":
    unittest.main()



"""Smartcard module exceptions.

This module defines the exceptions raised by the smartcard.pcsc modules.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

# gemalto scard library
import smartcard.scard


class BaseSCardException(Exception):
    """Base class for scard (aka PCSC) exceptions.

    scard exceptions are raised by the scard module, i.e.
    low-level PCSC access to readers and cards.

    """

    def __init__(self, hresult=-1, message="", *args):
        """Constructor that stores the pcsc error status."""
        if not message:
            message = "scard exception"
        super().__init__(message, *args)
        self.message = message
        self.hresult = hresult

    def __str__(self):
        """Returns a string representation of the exception."""
        text = super().__str__()
        if self.hresult != -1:
            hresult = self.hresult
            if hresult < 0:
                # convert 0x-7FEFFFE3 into 0x8010001D
                hresult += 0x100000000
            text += f": {smartcard.scard.SCardGetErrorMessage(self.hresult)} (0x{hresult:08X})"
        return text


class AddReaderToGroupException(BaseSCardException):
    """Raised when scard fails to add a new reader to a PCSC reader group."""

    def __init__(self, hresult, readername="", groupname=""):
        super().__init__(
            message="Failed to add reader: " + readername + " to group: " + groupname,
            hresult=hresult,
        )
        self.readername = readername
        self.groupname = groupname


class EstablishContextException(BaseSCardException):
    """Raised when scard failed to establish context with PCSC."""

    def __init__(self, hresult):
        super().__init__(message="Failed to establish context", hresult=hresult)


class ListReadersException(BaseSCardException):
    """Raised when scard failed to list readers."""

    def __init__(self, hresult):
        super().__init__(message="Failed to list readers", hresult=hresult)


class IntroduceReaderException(BaseSCardException):
    """Raised when scard fails to introduce a new reader to PCSC."""

    def __init__(self, hresult, readername=""):
        super().__init__(
            message="Failed to introduce a new reader: " + readername, hresult=hresult
        )
        self.readername = readername


class ReleaseContextException(BaseSCardException):
    """Raised when scard failed to release PCSC context."""

    def __init__(self, hresult):
        super().__init__(message="Failed to release context", hresult=hresult)


class RemoveReaderFromGroupException(BaseSCardException):
    """Raised when scard fails to remove a reader from a PCSC reader group."""

    def __init__(self, hresult, readername="", groupname=""):
        BaseSCardException.__init__(self, hresult)
        self.readername = readername
        self.groupname = groupname
        super().__init__(
            message="Failed to remove reader: "
            + readername
            + " from group: "
            + groupname,
            hresult=hresult,
        )


if __name__ == "__main__":
    try:
        raise EstablishContextException(smartcard.scard.SCARD_E_NO_MEMORY)
    except BaseSCardException as exc:
        print(exc)



#! /usr/bin/env python3
"""Unit tests for smartcard.Card

This test case can be executed individually, or with all other test cases
thru testsuite_framework.py.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""


import sys
import unittest

from smartcard.Card import Card
from smartcard.Exceptions import NoCardException
from smartcard.System import readers

sys.path += [".."]

try:
    from local_config import expectedATRinReader, expectedReaders
except ImportError:
    print("execute test suite first to generate the local_config.py file")
    sys.exit()


SELECT = [0xA0, 0xA4, 0x00, 0x00, 0x02]
DF_TELECOM = [0x7F, 0x10]


class testcase_Card(unittest.TestCase):
    """Test case for smartcard.Card."""

    def testcase_CardDictionary(self):
        """Create a dictionary with Card keys"""
        mydict = {}
        for reader in readers():
            card = Card(reader, expectedATRinReader[reader.name])
            mydict[card] = reader

        for card in list(mydict.keys()):
            self.assertTrue(str(card.reader) in expectedReaders)

    def testcase_Card_FromReaders(self):
        """Create a Card from Readers and test that the response
        to SELECT DF_TELECOM has two bytes."""

        for reader in readers():
            card = Card(reader, expectedATRinReader[reader.name])
            cc = card.createConnection()
            if expectedATRinReader[str(reader)]:
                cc.connect()
                response, sw1, sw2 = cc.transmit(SELECT + DF_TELECOM)
                expectedSWs = {"9f 1a": 1, "9f 20": 2, "6e 0": 3}
                self.assertEqual([], response)
                self.assertTrue(f"{sw1:x} {sw2:x}" in expectedSWs or "9f" == f"{sw1:x}")
            else:
                self.assertRaises(NoCardException, cc.connect)

    def testcase_Card_FromReaderStrings(self):
        """Create a Card from reader strings and test that the response
        to SELECT DF_TELECOM has two bytes."""

        for reader in readers():
            card = Card(str(reader), expectedATRinReader[reader.name])
            cc = card.createConnection()
            if expectedATRinReader[str(reader)]:
                cc.connect()
                response, sw1, sw2 = cc.transmit(SELECT + DF_TELECOM)
                expectedSWs = {"9f 1a": 1, "9f 20": 2, "6e 0": 3}
                self.assertEqual([], response)
                self.assertTrue(f"{sw1:x} {sw2:x}" in expectedSWs or "9f" == f"{sw1:x}")
            else:
                self.assertRaises(NoCardException, cc.connect)

    def testcase_Card_Eq_NotEq(self):
        """Test == and != for Cards."""
        for reader in readers():
            card = Card(str(reader), expectedATRinReader[reader.name])
            cardcopy = Card(str(reader), expectedATRinReader[str(reader)])
            self.assertEqual(True, card == cardcopy)
            self.assertEqual(True, not card != cardcopy)

        for reader in readers():
            card = Card(str(reader), expectedATRinReader[reader.name])
            cardcopy = Card(str(reader), [0, 0])
            self.assertEqual(True, card != cardcopy)
            self.assertEqual(True, not card == cardcopy)


if __name__ == "__main__":
    unittest.main(verbosity=1)



import pytest

from smartcard.util import *


def test_to_bytes():
    data_in = "3B 65 00 00 9C 11 01 01 03"
    data_out = [59, 101, 0, 0, 156, 17, 1, 1, 3]
    assert toBytes(data_in) == data_out

    data_in = "3B6500009C11010103"
    assert toBytes(data_in) == data_out

    data_in = "3B6500   009C1101  0103"
    assert toBytes(data_in) == data_out

    data_in = """
                3B 65 00
                00 9C 11 01
                01 03
              """
    assert toBytes(data_in) == data_out

    data_in = "zz"
    with pytest.raises(TypeError):
        toBytes(data_in)


def test_padd():
    data_in = toBytes("3B 65 00 00 9C 11 01 01 03")
    old_data_in = list(data_in)
    data_out = [
        0x3B,
        0x65,
        0,
        0,
        0x9C,
        0x11,
        1,
        1,
        3,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
        0xFF,
    ]
    assert padd(data_in, 16) == data_out
    assert data_in == old_data_in

    assert padd(data_in, 4) == data_in


def test_to_ascii_bytes():
    data_in = "Number 101"
    data_out = [0x4E, 0x75, 0x6D, 0x62, 0x65, 0x72, 0x20, 0x31, 0x30, 0x31]
    assert toASCIIBytes(data_in) == data_out


def test_to_ascii_string():
    data_in = [0x4E, 0x75, 0x6D, 0x62, 0x65, 0x72, 0x20, 0x31, 0x30, 0x31]
    data_out = "Number 101"
    assert toASCIIString(data_in) == data_out

    data_in = [0x01, 0x20, 0x80, 0x7E, 0xF0]
    data_out = ". .~."
    assert toASCIIString(data_in) == data_out


def test_to_gsm3_38_bytes():
    data_in = "@Pascal"
    data_out = [0x00, 0x50, 0x61, 0x73, 0x63, 0x61, 0x6C]
    assert toGSM3_38Bytes(data_in) == data_out

    data_in = "@ùPascal"
    data_out = [0x00, 0x06, 0x50, 0x61, 0x73, 0x63, 0x61, 0x6C]
    assert toGSM3_38Bytes(data_in) == data_out

    data_in = "@ùPascal".encode("iso8859-1")
    data_out = [0x00, 0x06, 0x50, 0x61, 0x73, 0x63, 0x61, 0x6C]
    assert toGSM3_38Bytes(data_in) == data_out

    data_in = "1234"
    data_out = [0x31, 0x32, 0x33, 0x34]
    assert toGSM3_38Bytes(data_in) == data_out


def test_to_hex_string():
    data_in = []
    data_out = ""
    assert toHexString(data_in) == data_out

    data_in = 42
    with pytest.raises(TypeError):
        toHexString(data_in)

    data_in = [0x3B, 0x65, 0x00, 0x00, 0x9C, 0x11, 0x01, 0x01, 0x03]
    data_out = "3B 65 00 00 9C 11 01 01 03"
    assert toHexString(data_in) == data_out

    data_out = "3B, 65, 00, 00, 9C, 11, 01, 01, 03"
    assert toHexString(data_in, COMMA) == data_out

    data_out = "0x3B 0x65 0x00 0x00 0x9C 0x11 0x01 0x01 0x03"
    assert toHexString(data_in, HEX) == data_out

    data_out = "0x3B, 0x65, 0x00, 0x00, 0x9C, 0x11, 0x01, 0x01, 0x03"
    assert toHexString(data_in, HEX | COMMA) == data_out

    data_out = "0X3B 0X65 0X00 0X00 0X9C 0X11 0X01 0X01 0X03"
    assert toHexString(data_in, HEX | UPPERCASE) == data_out

    data_out = "0X3B, 0X65, 0X00, 0X00, 0X9C, 0X11, 0X01, 0X01, 0X03"
    assert toHexString(data_in, HEX | UPPERCASE | COMMA) == data_out

    data_out = "3B6500009C11010103"
    assert toHexString(data_in, PACK) == data_out

    data_out = "3B,65,00,00,9C,11,01,01,03"
    assert toHexString(data_in, COMMA | PACK) == data_out

    data_out = "0x3B0x650x000x000x9C0x110x010x010x03"
    assert toHexString(data_in, HEX | PACK) == data_out

    data_out = "0x3B,0x65,0x00,0x00,0x9C,0x11,0x01,0x01,0x03"
    assert toHexString(data_in, HEX | COMMA | PACK) == data_out

    data_out = "0X3B0X650X000X000X9C0X110X010X010X03"
    assert toHexString(data_in, HEX | UPPERCASE | PACK) == data_out

    data_out = "0X3B,0X65,0X00,0X00,0X9C,0X11,0X01,0X01,0X03"
    assert toHexString(data_in, HEX | UPPERCASE | COMMA | PACK) == data_out

    with pytest.raises(TypeError):
        toHexString("foo")


def test_hex_list_to_bin_string():
    data_in = [1, 2, 3]
    data_out = "\x01\x02\x03"
    with pytest.warns(DeprecationWarning):
        assert HexListToBinString(data_in) == data_out


def test_bin_string_to_hex_list():
    data_in = "\x01\x02\x03"
    data_out = [1, 2, 3]
    with pytest.warns(DeprecationWarning):
        assert BinStringToHexList(data_in) == data_out


def test_hl2bs():
    data_in = [78, 117, 109, 98, 101, 114, 32, 49, 48, 49]
    data_out = "Number 101"
    with pytest.warns(DeprecationWarning):
        assert hl2bs(data_in) == data_out


def test_bs2hl():
    data_in = "Number 101"
    data_out = [78, 117, 109, 98, 101, 114, 32, 49, 48, 49]
    with pytest.warns(DeprecationWarning):
        assert bs2hl(data_in) == data_out



"""
from Thinking in Python, Bruce Eckel
https://python-3-patterns-idioms-test.readthedocs.io/en/latest/Observer.html

(c) Copyright 2008, Creative Commons Attribution-Share Alike 3.0.

Simple emulation of Java's 'synchronized'
keyword, from Peter Norvig.
"""

from __future__ import annotations

import functools
import sys
import threading
from collections.abc import Iterable
from typing import Any, Callable, Protocol, TypeVar

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec


T = TypeVar("T")
P = ParamSpec("P")


def synchronized(method: Callable[P, T]) -> Callable[P, T]:
    @functools.wraps(method)
    def f(self: _SynchronizationProtocol, *args: Any, **kwargs: Any) -> Any:
        with self.mutex:
            return method(self, *args, **kwargs)

    return f


def synchronize(klass: type, names: str | Iterable[str] | None = None) -> None:
    """Synchronize methods in the given class.
    Only synchronize the methods whose names are
    given, or all methods if names=None."""

    if isinstance(names, str):
        names = names.split()
    for name, val in list(klass.__dict__.items()):
        if callable(val) and name != "__init__" and (names is None or name in names):
            setattr(klass, name, synchronized(val))


class _SynchronizationProtocol(Protocol):
    mutex: threading.Lock | threading.RLock


class Synchronization(_SynchronizationProtocol):
    # You can create your own self.mutex, or inherit from this class:

    def __init__(self):
        self.mutex = threading.RLock()



#! /usr/bin/env python3
"""Sample script for APDU error checking with a custom error checker.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from smartcard.CardConnectionObserver import ConsoleCardConnectionObserver
from smartcard.CardRequest import CardRequest
from smartcard.CardType import AnyCardType
from smartcard.sw.ErrorChecker import ErrorChecker
from smartcard.sw.ErrorCheckingChain import ErrorCheckingChain
from smartcard.sw.SWExceptions import SWException


class MyErrorChecker(ErrorChecker):
    """Our custom error checker that will except if 0x61<sw1<0x70."""

    def __call__(self, data, sw1, sw2):
        print(sw1, sw2)
        if 0x61 < sw1 and 0x70 > sw1:
            raise SWException(data, sw1, sw2)


# define the apdus used in this script
GET_RESPONSE = [0xA0, 0xC0, 00, 00]
SELECT = [0xA0, 0xA4, 0x00, 0x00, 0x02]
DF_TELECOM = [0x7F, 0x10]

if __name__ == "__main__":

    print("Insert a card within 10 seconds")
    print("Cards without a DF_TELECOM will except")

    # request any card
    cardtype = AnyCardType()
    cardrequest = CardRequest(timeout=10, cardType=cardtype)
    cardservice = cardrequest.waitforcard()

    # our error checking chain
    errorchain = []
    errorchain = [ErrorCheckingChain([], MyErrorChecker())]
    cardservice.connection.setErrorCheckingChain(errorchain)

    # attach the console tracer
    observer = ConsoleCardConnectionObserver()
    cardservice.connection.addObserver(observer)

    # send a few apdus; exceptions will occur upon errors
    cardservice.connection.connect()

    try:
        SELECT = [0xA0, 0xA4, 0x00, 0x00, 0x02]
        DF_TELECOM = [0x7F, 0x10]
        apdu = SELECT + DF_TELECOM
        response, sw1, sw2 = cardservice.connection.transmit(apdu)
        if sw1 == 0x9F:
            GET_RESPONSE = [0xA0, 0xC0, 00, 00]
            apdu = GET_RESPONSE + [sw2]
            response, sw1, sw2 = cardservice.connection.transmit(apdu)
    except SWException as e:
        print(e, f"{e.sw1:x} {e.sw2:x}")

    cardservice.connection.disconnect()
    cardservice.connection.release()

    import sys

    if "win32" == sys.platform:
        print("press Enter to continue")
        sys.stdin.read(1)



#! /usr/bin/env python3
"""Unit tests for smartcard.CardService

This test case can be executed individually, or with all other test cases
thru testsuite_framework.py.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import sys
import unittest

from smartcard.CardService import CardService
from smartcard.System import readers

sys.path += [".."]

try:
    from local_config import expectedATRinReader
except ImportError:
    print("execute test suite first to generate the local_config.py file")
    sys.exit()


class testcase_CardService(unittest.TestCase):
    """Test case for CardService."""

    def testcase_CardService(self):
        """Test the response to SELECT DF_TELECOM."""
        SELECT = [0xA0, 0xA4, 0x00, 0x00, 0x02]
        DF_TELECOM = [0x7F, 0x10]

        for reader in readers():
            if expectedATRinReader[reader.name]:
                cc = reader.createConnection()
                cs = CardService(cc)
                cs.connection.connect()
                response, sw1, sw2 = cs.connection.transmit(SELECT + DF_TELECOM)
                expectedSWs = {"9f 1a": 1, "6e 0": 2, "9f 20": 3, "9f 22": 4}
                self.assertEqual([], response)
                self.assertTrue(f"{sw1:x} {sw2:x}" in expectedSWs or "9f" == f"{sw1:x}")


if __name__ == "__main__":
    unittest.main(verbosity=1)



#! /usr/bin/env python3
"""
Sample for python PCSC wrapper module: send a Control Code to a card or
reader

__author__ = "Ludovic Rousseau"

Copyright 2009-2010 Ludovic Rousseau
Author: Ludovic Rousseau, mailto:ludovic.rousseau@free.fr

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import sys

from smartcard.pcsc.PCSCExceptions import *
from smartcard.scard import *


def can_do_verify_pin(hCard):
    FEATURE_VERIFY_PIN_DIRECT = 6
    return parse_get_feature_request(hCard, FEATURE_VERIFY_PIN_DIRECT)


def can_do_modify_pin(hCard):
    FEATURE_MODIFY_PIN_DIRECT = 7
    return parse_get_feature_request(hCard, FEATURE_MODIFY_PIN_DIRECT)


def parse_get_feature_request(hCard, feature):
    # check the reader can do a verify pin
    CM_IOCTL_GET_FEATURE_REQUEST = SCARD_CTL_CODE(3400)
    hresult, response = SCardControl(hcard, CM_IOCTL_GET_FEATURE_REQUEST, [])
    if hresult != SCARD_S_SUCCESS:
        raise BaseSCardException(hresult)
    print(response)
    while len(response) > 0:
        tag = response[0]
        if feature == tag:
            return (
                ((((response[2] << 8) + response[3]) << 8) + response[4]) << 8
            ) + response[5]
        response = response[6:]


def verifypin(hCard, control=None):
    if control is None:
        control = can_do_verify_pin(hCard)
        if control is None:
            raise Exception("Not a pinpad")

    command = [
        0x00,  # bTimerOut
        0x00,  # bTimerOut2
        0x82,  # bmFormatString
        0x04,  # bmPINBlockString
        0x00,  # bmPINLengthFormat
        0x08,
        0x04,  # wPINMaxExtraDigit
        0x02,  # bEntryValidationCondition
        0x01,  # bNumberMessage
        0x04,
        0x09,  # wLangId
        0x00,  # bMsgIndex
        0x00,
        0x00,
        0x00,  # bTeoPrologue
        13,
        0,
        0,
        0,  # ulDataLength
        0x00,
        0x20,
        0x00,
        0x00,
        0x08,
        0x30,
        0x30,
        0x30,
        0x30,
        0x30,
        0x30,
        0x30,
        0x30,
    ]  # abData
    hresult, response = SCardControl(hcard, control, command)
    if hresult != SCARD_S_SUCCESS:
        raise BaseSCardException(hresult)
    return hresult, response


try:
    hresult, hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
    if hresult != SCARD_S_SUCCESS:
        raise EstablishContextException(hresult)
    print("Context established!")

    try:
        hresult, readers = SCardListReaders(hcontext, [])
        if hresult != SCARD_S_SUCCESS:
            raise ListReadersException(hresult)
        print("PCSC Readers:", readers)

        if len(readers) < 1:
            raise Exception("No smart card readers")

        for zreader in readers:

            print("Trying to Control reader:", zreader)

            try:
                hresult, hcard, dwActiveProtocol = SCardConnect(
                    hcontext,
                    zreader,
                    SCARD_SHARE_SHARED,
                    SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1,
                )
                if hresult != SCARD_S_SUCCESS:
                    raise BaseSCardException(hresult)
                print("Connected with active protocol", dwActiveProtocol)

                try:
                    SELECT = [
                        0x00,
                        0xA4,
                        0x04,
                        0x00,
                        0x06,
                        0xA0,
                        0x00,
                        0x00,
                        0x00,
                        0x18,
                        0xFF,
                    ]
                    hresult, response = SCardTransmit(hcard, dwActiveProtocol, SELECT)
                    if hresult != SCARD_S_SUCCESS:
                        raise BaseSCardException(hresult)

                    cmd_verify = can_do_verify_pin(hcard)
                    if cmd_verify:
                        print("can do verify pin: 0x%08X" % cmd_verify)

                    cmd_modify = can_do_modify_pin(hcard)
                    if cmd_modify:
                        print("can do modify pin: 0x%08X" % cmd_modify)

                    hresult, response = verifypin(hcard, cmd_verify)
                    print("Control:", response)
                except Exception as ex:
                    print("Exception:", ex)
                finally:
                    hresult = SCardDisconnect(hcard, SCARD_UNPOWER_CARD)
                    if hresult != SCARD_S_SUCCESS:
                        raise BaseSCardException(hresult)
                    print("Disconnected")

            except BaseSCardException as ex:
                print("SCard Exception:", ex)

    finally:
        hresult = SCardReleaseContext(hcontext)
        if hresult != SCARD_S_SUCCESS:
            raise ReleaseContextException(hresult)
        print("Released context.")

except error as e:
    print(e)

if "win32" == sys.platform:
    print("press Enter to continue")
    sys.stdin.read(1)



#! /usr/bin/env python3
"""Unit tests for ulist

This test case can be executed individually, or with all other test cases
thru testsuite_framework.py.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import unittest

from smartcard.ulist import ulist


class C(ulist):
    """ulist subclass"""

    def __onadditem__(self, item):
        # print('+', item)
        pass

    def __onremoveitem__(self, item):
        # print('-', item)
        pass


class testcase_ulist(unittest.TestCase):
    """Test smartcard.ulist."""

    def testcase_ulist_init(self):
        """tests constructor"""

        c = C([1, 2, 3, 3, 4, 5, 5])
        self.assertEqual([1, 2, 3, 4, 5], c)

        c = C(["one", "two", "three", "one"])
        self.assertEqual(["one", "two", "three"], c)

    def testcase_ulist_add(self):
        """tests l=l+other"""

        seed = [1, 2, 3]
        c = C(seed)
        self.assertEqual(seed, c)

        c = c + []
        self.assertEqual(seed, c)

        c = c + 4
        self.assertEqual(seed + [4], c)

        c = c + 4
        self.assertEqual(seed + [4], c)

        c = c + "word"
        self.assertEqual(seed + [4] + ["word"], c)

        seed = ["one", "two", "three"]
        c = C(seed)
        self.assertEqual(seed, c)

        c = c + ["four", "five"]
        self.assertEqual(seed + ["four", "five"], c)

    def testcase_ulist_iadd(self):
        """tests l+=other"""

        seed = [1, 2, 3]
        c = C(seed)
        self.assertEqual(seed, c)

        c += []
        self.assertEqual(seed, c)

        c += 4
        self.assertEqual(seed + [4], c)

        c += 4
        self.assertEqual(seed + [4], c)

        c += [4, 3, 2, 1]
        self.assertEqual(seed + [4], c)

        c += "word"
        self.assertEqual(seed + [4] + ["word"], c)

        seed = ["one", "two", "three"]
        c = C(seed)
        self.assertEqual(seed, c)

        c += ["four", "five"]
        self.assertEqual(seed + ["four", "five"], c)

    def testcase_ulist_radd(self):
        """tests l=other+l"""

        seed = [1, 2, 3]
        c = C(seed)
        self.assertEqual(seed, c)

        l = [] + c
        self.assertEqual(seed, l)

        l = [3] + c
        self.assertEqual(seed, c)
        self.assertEqual(seed, l)

        l = [3, 3, 4, 4] + c
        self.assertEqual(seed, c)
        self.assertEqual(seed + [4], l)

        l = [4] + ["word"] + c
        self.assertEqual(seed, c)
        self.assertEqual(seed + [4] + ["word"], l)

    def testcase_ulist_append(self):
        """append"""

        seed = [1, 2, 3]
        c = C(seed)

        c.append(4)
        self.assertEqual(seed + [4], c)

        c.append(4)
        self.assertEqual(seed + [4], c)

        c.append("word")
        self.assertEqual(seed + [4] + ["word"], c)

    def testcase_ulist_insert(self):
        """insert"""

        seed = [1, 2, 3]
        c = C(seed)

        c.insert(0, 0)
        self.assertEqual([0] + seed, c)

        c.insert(1, 0)
        self.assertEqual([0] + seed, c)

    def testcase_ulist_pop(self):
        """pop"""

        seed = [1, 2, 3]
        c = C(seed)

        c.pop()
        self.assertEqual(c, [1, 2])

        c.pop(1)
        self.assertEqual(c, [1])

    def testcase_ulist_remove(self):
        """remove"""

        seed = [1, 2, 3]
        c = C(seed)

        c.remove(2)
        self.assertEqual(c, [1, 3])

        c.remove(1)
        self.assertEqual(c, [3])


if __name__ == "__main__":
    unittest.main(verbosity=1)



"""ulist is a subclass of list where items cannot appear twice in the list.

[1,2,2,3,3,4] is a valid list, whereas in ulist we can only have [1,2,3,4].

__author__ = "gemalto https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""


class ulist(list):
    """ulist ensures that all items are unique and provides an __onadditem__
    hook to perform custom action in subclasses."""

    #
    # override list methods
    #

    def __init__(self, initlist=None):
        if initlist is not None and initlist != []:
            list.__init__(self, [initlist[0]])
            for item in initlist[1:]:
                if not list.__contains__(self, item):
                    list.append(self, item)
        else:
            list.__init__(self, initlist)

    def __add__(self, other):
        newother = self.__remove_duplicates(other)
        self.__appendother__(newother)
        return self.__class__(list(self) + list(newother))

    def __iadd__(self, other):
        newother = self.__remove_duplicates(other)
        self.__appendother__(newother)
        list.__iadd__(self, newother)
        return self

    def __radd__(self, other):
        newother = self.__remove_duplicates(other)
        return list.__add__(self, newother)

    def append(self, item):
        if not list.__contains__(self, item):
            list.append(self, item)
            self.__onadditem__(item)

    def insert(self, i, item):
        if not list.__contains__(self, item):
            list.insert(self, i, item)
            self.__onadditem__(item)

    def pop(self, i=-1):
        item = list.pop(self, i)
        self.__onremoveitem__(item)
        return item

    def remove(self, item):
        list.remove(self, item)
        self.__onremoveitem__(item)

    #
    # non list methods
    #

    def __remove_duplicates(self, _other):
        """Remove from other items already in list."""
        if (
            not isinstance(_other, type(self))
            and not isinstance(_other, type(list))
            and not isinstance(_other, list)
        ):
            other = [_other]
        else:
            other = list(_other)

        # remove items already in self
        newother = []
        for i in range(0, len(other)):
            item = other.pop(0)
            if not list.__contains__(self, item):
                newother.append(item)

        # remove duplicate items in other
        other = []
        if newother != []:
            other.append(newother[0])
            for i in range(1, len(newother)):
                item = newother.pop()
                if not other.__contains__(item):
                    other.append(item)
        return other

    def __appendother__(self, other):
        """Append other to object."""
        for item in other:
            self.__onadditem__(item)

    def __onadditem__(self, item):
        """Called for each item added. Override in subclasses for adding
        custom action."""
        pass

    def __onremoveitem__(self, item):
        """Called for each item removed. Override in subclasses for
        adding custom action."""
        pass



#! /usr/bin/env python3
"""Unit tests for testing that we have at least two cards for the test suite.

This test case can be executed individually, or with all other test cases
thru testsuite_framework.py.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import sys
import unittest

sys.path += [".."]

try:
    from local_config import expectedATRs
except ImportError:
    print("execute test suite first to generate the local_config.py file")
    sys.exit()


class testcase_ATR(unittest.TestCase):
    """Test ATR configuration for test suite"""

    def testcase_ATRconfig(self):
        """ATRconfig"""
        # we have at list 2 readers (one is the simulator), e.g.
        # two potential ATRs
        self.assertTrue(len(expectedATRs) > 1)

        # we have at least two non empty ATRs
        count = 0
        for atr in expectedATRs:
            if atr:
                count += 1
        self.assertTrue(count > 1)


if __name__ == "__main__":
    unittest.main(verbosity=1)



"""PCSCCardConnection class manages connections thru a PCSC reader.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from smartcard.CardConnection import CardConnection
from smartcard.Exceptions import (
    CardConnectionException,
    NoCardException,
    SmartcardException,
)
from smartcard.scard import *


def translateprotocolmask(protocol):
    """Translate L{CardConnection} protocol mask into PCSC protocol mask."""
    pcscprotocol = 0
    if protocol is not None:
        if CardConnection.T0_protocol & protocol:
            pcscprotocol |= SCARD_PROTOCOL_T0
        if CardConnection.T1_protocol & protocol:
            pcscprotocol |= SCARD_PROTOCOL_T1
        if CardConnection.RAW_protocol & protocol:
            pcscprotocol |= SCARD_PROTOCOL_RAW
        if CardConnection.T15_protocol & protocol:
            pcscprotocol |= SCARD_PROTOCOL_T15
    return pcscprotocol


def translateprotocolheader(protocol):
    """Translate protocol into PCSC protocol header."""
    pcscprotocol = 0
    if protocol is not None:
        if CardConnection.T0_protocol == protocol:
            pcscprotocol = SCARD_PCI_T0
        if CardConnection.T1_protocol == protocol:
            pcscprotocol = SCARD_PCI_T1
        if CardConnection.RAW_protocol == protocol:
            pcscprotocol = SCARD_PCI_RAW
    return pcscprotocol


dictProtocolHeader = {SCARD_PCI_T0: "T0", SCARD_PCI_T1: "T1", SCARD_PCI_RAW: "RAW"}
dictProtocol = {
    SCARD_PROTOCOL_T0: "T0",
    SCARD_PROTOCOL_T1: "T1",
    SCARD_PROTOCOL_RAW: "RAW",
    SCARD_PROTOCOL_T15: "T15",
    SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1: "T0 or T1",
}


class PCSCCardConnection(CardConnection):
    """PCSCCard connection class. Handles connection with a card thru a
    PCSC reader."""

    def __init__(self, reader):
        """Construct a new PCSC card connection.

        @param reader: the reader in which the smartcard to connect to is located.
        """
        CardConnection.__init__(self, reader)
        self.hcard = None
        hresult, self.hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
        if hresult != SCARD_S_SUCCESS:
            raise CardConnectionException(
                "Failed to establish context : " + SCardGetErrorMessage(hresult),
                hresult=hresult,
            )

    def __del__(self):
        """Destructor. Clean PCSC connection resources."""
        # race condition: module CardConnection
        # can disappear before __del__ is called
        self.release()

    def release(self):
        """explicit release"""
        if self.hcontext is not None:
            CardConnection.release(self)
            self.disconnect()
            hresult = SCardReleaseContext(self.hcontext)
            if hresult != SCARD_S_SUCCESS and hresult != SCARD_E_INVALID_VALUE:
                raise CardConnectionException(
                    "Failed to release context: " + SCardGetErrorMessage(hresult),
                    hresult=hresult,
                )
            self.hcontext = None
        CardConnection.__del__(self)

    def connect(self, protocol=None, mode=None, disposition=None):
        """Connect to the card.

        If protocol is not specified, connect with the default
        connection protocol.

        If mode is not specified, connect with
        C{smartcard.scard.SCARD_SHARE_SHARED}."""
        CardConnection.connect(self, protocol)
        pcscprotocol = translateprotocolmask(protocol)
        if 0 == pcscprotocol:
            pcscprotocol = self.getProtocol()

        if mode is None:
            mode = SCARD_SHARE_SHARED

        # store the way to dispose the card
        if disposition is None:
            disposition = SCARD_UNPOWER_CARD
        self.disposition = disposition

        hresult, self.hcard, dwActiveProtocol = SCardConnect(
            self.hcontext, str(self.reader), mode, pcscprotocol
        )
        if hresult != SCARD_S_SUCCESS:
            self.hcard = None
            if hresult in (SCARD_W_REMOVED_CARD, SCARD_E_NO_SMARTCARD):
                raise NoCardException("Unable to connect", hresult=hresult)
            else:
                raise CardConnectionException(
                    "Unable to connect with protocol: "
                    + dictProtocol[pcscprotocol]
                    + ". "
                    + SCardGetErrorMessage(hresult),
                    hresult=hresult,
                )

        protocol = 0
        if dwActiveProtocol == SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1:
            # special case for T0 | T1
            # this happens when mode=SCARD_SHARE_DIRECT and no protocol is
            # then negotiated with the card
            protocol = CardConnection.T0_protocol | CardConnection.T1_protocol
        else:
            for p in dictProtocol:
                if p == dwActiveProtocol:
                    protocol = eval("CardConnection.%s_protocol" % dictProtocol[p])
        PCSCCardConnection.setProtocol(self, protocol)

    def reconnect(self, protocol=None, mode=None, disposition=None):
        """Reconnect to the card.

        If protocol is not specified, connect with the default
        connection protocol.

        If mode is not specified, connect with
        C{smartcard.scard.SCARD_SHARE_SHARED}.

        If disposition is not specified, do a warm reset
        (C{smartcard.scard.SCARD_RESET_CARD})"""
        CardConnection.reconnect(self, protocol)
        if self.hcard is None:
            raise CardConnectionException("Card not connected")

        pcscprotocol = translateprotocolmask(protocol)
        if 0 == pcscprotocol:
            pcscprotocol = self.getProtocol()

        if mode is None:
            mode = SCARD_SHARE_SHARED

        # store the way to dispose the card
        if disposition is None:
            disposition = SCARD_RESET_CARD
        self.disposition = disposition

        hresult, dwActiveProtocol = SCardReconnect(
            self.hcard, mode, pcscprotocol, self.disposition
        )
        if hresult != SCARD_S_SUCCESS:
            self.hcard = None
            if hresult in (SCARD_W_REMOVED_CARD, SCARD_E_NO_SMARTCARD):
                raise NoCardException("Unable to reconnect", hresult=hresult)
            else:
                raise CardConnectionException(
                    "Unable to reconnect with protocol: "
                    + dictProtocol[pcscprotocol]
                    + ". "
                    + SCardGetErrorMessage(hresult),
                    hresult=hresult,
                )

        protocol = 0
        if dwActiveProtocol == SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1:
            # special case for T0 | T1
            # this happens when mode=SCARD_SHARE_DIRECT and no protocol is
            # then negotiated with the card
            protocol = CardConnection.T0_protocol | CardConnection.T1_protocol
        else:
            for p in dictProtocol:
                if p == dwActiveProtocol:
                    protocol = eval("CardConnection.%s_protocol" % dictProtocol[p])
        PCSCCardConnection.setProtocol(self, protocol)

    def disconnect(self):
        """Disconnect from the card."""

        if self.hcard is not None:
            # when __del__() is invoked in response to a module being
            # deleted, e.g., when execution of the program is done,
            # other globals referenced by the __del__() method may
            # already have been deleted.  this causes
            # CardConnection.disconnect to except with a TypeError
            try:
                CardConnection.disconnect(self)
            except TypeError:
                pass
            hresult = SCardDisconnect(self.hcard, self.disposition)
            self.hcard = None
            if hresult != SCARD_S_SUCCESS:
                raise CardConnectionException(
                    "Failed to disconnect: " + SCardGetErrorMessage(hresult),
                    hresult=hresult,
                )

    def getATR(self):
        """Return card ATR"""
        CardConnection.getATR(self)
        if self.hcard is None:
            raise CardConnectionException("Card not connected")
        hresult, reader, state, protocol, atr = SCardStatus(self.hcard)
        if hresult != SCARD_S_SUCCESS:
            raise CardConnectionException(
                "Failed to get status: " + SCardGetErrorMessage(hresult),
                hresult=hresult,
            )
        return atr

    def doTransmit(self, command, protocol=None):
        """Transmit an apdu to the card and return response apdu.

        @param command:  command apdu to transmit (list of bytes)

        @param protocol: the transmission protocol, from
            L{CardConnection.T0_protocol}, L{CardConnection.T1_protocol}, or
            L{CardConnection.RAW_protocol}

        @return:     a tuple (response, sw1, sw2) where
                    - response are the response bytes excluding status words
                    - sw1 is status word 1, e.g. 0x90
                    - sw2 is status word 2, e.g. 0x1A
        """
        if protocol is None:
            protocol = self.getProtocol()
        CardConnection.doTransmit(self, command, protocol)
        pcscprotocolheader = translateprotocolheader(protocol)
        if 0 == pcscprotocolheader:
            raise CardConnectionException(
                "Invalid protocol in transmit: must be "
                + "CardConnection.T0_protocol, "
                + "CardConnection.T1_protocol, or "
                + "CardConnection.RAW_protocol"
            )
        if self.hcard is None:
            raise CardConnectionException("Card not connected")
        hresult, response = SCardTransmit(self.hcard, pcscprotocolheader, command)
        if hresult != SCARD_S_SUCCESS:
            raise CardConnectionException(
                "Failed to transmit with protocol "
                + dictProtocolHeader[pcscprotocolheader]
                + ". "
                + SCardGetErrorMessage(hresult),
                hresult=hresult,
            )

        if len(response) < 2:
            raise CardConnectionException(
                "Card returned no valid response", hresult=hresult
            )

        sw1 = (response[-2] + 256) % 256
        sw2 = (response[-1] + 256) % 256

        data = [(x + 256) % 256 for x in response[:-2]]
        return list(data), sw1, sw2

    def doControl(self, controlCode, command=None):
        """Transmit a control command to the reader and return response.

        @param controlCode: control command

        @param command:     command data to transmit (list of bytes)

        @return:      response are the response bytes (if any)
        """
        if command is None:
            command = []
        CardConnection.doControl(self, controlCode, command)
        hresult, response = SCardControl(self.hcard, controlCode, command)
        if hresult != SCARD_S_SUCCESS:
            raise SmartcardException(
                "Failed to control " + SCardGetErrorMessage(hresult), hresult=hresult
            )

        data = [(x + 256) % 256 for x in response]
        return list(data)

    def doGetAttrib(self, attribId):
        """get an attribute

        @param attribId: Identifier for the attribute to get

        @return:   response are the attribute byte array
        """
        CardConnection.doGetAttrib(self, attribId)
        hresult, response = SCardGetAttrib(self.hcard, attribId)
        if hresult != SCARD_S_SUCCESS:
            raise SmartcardException(
                "Failed to getAttrib " + SCardGetErrorMessage(hresult), hresult=hresult
            )
        return response


if __name__ == "__main__":
    """Small sample illustrating the use of CardConnection."""
    SELECT = [0xA0, 0xA4, 0x00, 0x00, 0x02]
    DF_TELECOM = [0x7F, 0x10]
    from smartcard.pcsc.PCSCPart10 import CM_IOCTL_GET_FEATURE_REQUEST
    from smartcard.pcsc.PCSCReader import PCSCReader

    cc = PCSCReader.readers()[0].createConnection()
    cc.connect()
    print("%r %x %x" % cc.transmit(SELECT + DF_TELECOM))

    print(cc.control(CM_IOCTL_GET_FEATURE_REQUEST, []))



import pytest

import smartcard.reader.ReaderGroups


@pytest.fixture(autouse=True)
def reset_reader_groups():
    """Reader group instances must be reset at the beginning of each test."""

    smartcard.reader.ReaderGroups.readergroups.instance = None
    yield



#! /usr/bin/env python3
"""Setup file for setuptools

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

Copyright 2007-2018 Ludovic Rousseau ludovic.rousseau@free.fr

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import platform
import shlex
import subprocess
from shutil import which
from sysconfig import get_platform

from setuptools import Extension, setup
from setuptools.command.build_py import build_py

platform_include_dirs = []
platform_sources = []
platform_libraries = []
platform_extra_compile_args = []  # ['-ggdb', '-O0']
platform_extra_link_args = []  # ['-ggdb']

if platform.system() == "Windows":
    platform__cc_defines = [("WIN32", "100")]
    platform_swig_opts = ["-DWIN32"]
    if "mingw" not in get_platform():
        platform_sources = ["src/smartcard/scard/scard.rc"]
    platform_libraries = ["winscard"]

elif platform.system() == "Darwin":
    platform__cc_defines = [("PCSCLITE", "1"), ("__APPLE__", "1")]
    platform_swig_opts = ["-DPCSCLITE", "-D__APPLE__"]

# Other (GNU/Linux, etc.)
#
else:
    platform__cc_defines = [("PCSCLITE", "1")]
    platform_swig_opts = ["-DPCSCLITE"]
    try:
        pkg_config_cflags = subprocess.check_output(
            ["pkg-config", "--cflags", "libpcsclite"]
        )
        platform_extra_compile_args += shlex.split(pkg_config_cflags.decode())
    except:
        platform_include_dirs = ["/usr/include/PCSC", "/usr/local/include/PCSC"]

VERSION_INFO = (2, 2, 2, 0)
VERSION_STR = "%i.%i.%i" % VERSION_INFO[:3]
VERSION_ALT = "%i,%01i,%01i,%04i" % VERSION_INFO


class BuildPyBuildExtFirst(build_py):
    """Workaround substitute `build_py` command for SWIG"""

    def run(self):
        if which("swig") is None:
            print("Install swig and try again")
            print("")
            exit(1)
        # Run build_ext first so that SWIG generated files are included
        self.run_command("build_ext")
        return build_py.run(self)


kw = {
    "name": "pyscard",
    "version": VERSION_STR,
    "description": "Smartcard module for Python.",
    "author": "Ludovic Rousseau",
    "author_email": "ludovic.rousseau@free.fr",
    "url": "https://github.com/LudovicRousseau/pyscard",
    "long_description": "Smartcard package for Python",
    "platforms": ["linux", "win32"],
    "python_requires": ">=3.9",
    "packages": [
        "smartcard",
        "smartcard.pcsc",
        "smartcard.reader",
        "smartcard.scard",
        "smartcard.sw",
        "smartcard.util",
        "smartcard.wx",
    ],
    "package_dir": {"": "src"},
    "package_data": {
        "smartcard.wx": ["resources/*.ico"],
    },
    "cmdclass": {"build_py": BuildPyBuildExtFirst},
    # the _scard.pyd extension to build
    "ext_modules": [
        Extension(
            "smartcard.scard._scard",
            define_macros=[
                ("VER_PRODUCTVERSION", VERSION_ALT),
                ("VER_PRODUCTVERSION_STR", VERSION_STR),
            ]
            + platform__cc_defines,
            include_dirs=["src/smartcard/scard/"] + platform_include_dirs,
            sources=[
                "src/smartcard/scard/helpers.c",
                "src/smartcard/scard/winscarddll.c",
                "src/smartcard/scard/scard.i",
            ]
            + platform_sources,
            libraries=platform_libraries,
            extra_compile_args=platform_extra_compile_args,
            extra_link_args=platform_extra_link_args,
            swig_opts=["-outdir", "src/smartcard/scard"] + platform_swig_opts,
        )
    ],
    "install_requires": [
        "typing_extensions; python_version=='3.9'",
    ],
    "extras_require": {
        "Gui": ["wxPython"],
    },
    "classifiers": [
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
        "Intended Audience :: Developers",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Security",
    ],
}

# FIXME Sourceforge downloads are unauthenticated, migrate to PyPI
kw["download_url"] = (
    "https://sourceforge.net/projects/%(name)s/files"
    "/%(name)s/%(name)s%%20%(version)s"
    "/%(name)s-%(version)s.tar.gz/download" % kw
)

setup(**kw)



#! /usr/bin/env python3
"""Unit tests for SCardBeginTransaction/SCardEndTransaction.

This test case can be executed individually, or with all other test cases
thru testsuite_scard.py.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""


# import local_config for reader/card configuration
# configcheck.py is generating local_config.py in
# the test suite.
import sys
import unittest

from smartcard.scard import *

sys.path += [".."]

try:
    from local_config import expectedATRs, expectedReaders
except ImportError:
    print("execute test suite first to generate the local_config.py file")
    sys.exit()


class testcase_transaction(unittest.TestCase):
    """Test scard API for SCardBegin/EndTransaction"""

    def setUp(self):
        hresult, self.hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
        self.assertEqual(hresult, 0)
        hresult, self.readers = SCardListReaders(self.hcontext, [])
        self.assertEqual(hresult, 0)

    def tearDown(self):
        hresult = SCardReleaseContext(self.hcontext)
        self.assertEqual(hresult, 0)

    def _transaction(self, r):
        if r < len(expectedATRs) and [] != expectedATRs[r]:
            hresult, hcard, dwActiveProtocol = SCardConnect(
                self.hcontext,
                self.readers[r],
                SCARD_SHARE_SHARED,
                SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1,
            )
            self.assertEqual(hresult, 0)

            try:
                hresult = SCardBeginTransaction(hcard)
                self.assertEqual(hresult, 0)

                hresult, reader, state, protocol, atr = SCardStatus(hcard)
                self.assertEqual(hresult, 0)
                self.assertEqual(reader, expectedReaders[r])
                self.assertEqual(atr, expectedATRs[r])

                hresult = SCardEndTransaction(hcard, SCARD_LEAVE_CARD)
                self.assertEqual(hresult, 0)

            finally:
                hresult = SCardDisconnect(hcard, SCARD_UNPOWER_CARD)
                self.assertEqual(hresult, 0)

    def test_transaction_reader0(self):
        testcase_transaction._transaction(self, 0)

    def test_transaction_reader1(self):
        testcase_transaction._transaction(self, 1)

    def test_transaction_reader2(self):
        testcase_transaction._transaction(self, 2)

    def test_transaction_reader3(self):
        testcase_transaction._transaction(self, 3)


def suite():
    suite1 = unittest.defaultTestLoader.loadTestsFromTestCase(testcase_transaction)
    return unittest.TestSuite(suite1)


if __name__ == "__main__":
    unittest.main()



import platform

import pytest

from smartcard.Exceptions import *
from smartcard.scard import *


def test_hresult_value():
    exc = SmartcardException()
    assert exc.hresult == -1


def test_list_readers_exception():
    exc = ListReadersException(-1)
    assert str(exc) == "Failed to list readers"

    exc = ListReadersException(0)
    assert exc.hresult == 0
    text = str(exc)
    if platform.system() == "Windows":
        expected = (
            "Failed to list readers: "
            "The operation completed successfully.  (0x00000000)"
        )
    else:
        expected = "Failed to list readers: Command successful. (0x00000000)"
    assert text == expected

    exc = ListReadersException(0x42)
    assert exc.hresult == 0x42
    text = str(exc)
    if platform.system() != "Windows":
        expected = "Failed to list readers: Unknown error: 0x00000042 (0x00000042)"
        macos_bug_expected = expected.replace("Unknown", "Unkown")
        assert text in (expected, macos_bug_expected)

    exc = ListReadersException(SCARD_S_SUCCESS)
    assert exc.hresult == 0

    exc = ListReadersException(SCARD_E_NO_SERVICE)
    assert exc.hresult == SCARD_E_NO_SERVICE
    text = str(exc)
    if platform.system() == "Windows":
        expected = (
            "Failed to list readers: "
            "The Smart Card Resource Manager is not running.  (0x8010001D)"
        )
    else:
        expected = "Failed to list readers: Service not available. (0x8010001D)"

    assert text == expected

    exc = ListReadersException(SCARD_E_NOT_TRANSACTED)
    if platform.system() == "Windows":
        expected = (
            "Failed to list readers: "
            "An attempt was made to end a non-existent transaction.  (0x80100016)"
        )
    else:
        expected = "Failed to list readers: Transaction failed. (0x80100016)"
    assert str(exc) == expected


def test_no_readers_exception():
    exc = NoReadersException()
    assert exc.hresult == -1
    text = str(exc)
    assert text == "No reader found"


def test_invalid_reader_exception():
    exc = InvalidReaderException("foobar")
    assert exc.hresult == -1
    text = str(exc)
    assert text == "Invalid reader: foobar"


def test_card_connection_exception():
    exc = CardConnectionException()
    assert exc.hresult == -1
    text = str(exc)
    assert text == ""

    exc = CardConnectionException(hresult=SCARD_W_REMOVED_CARD)
    assert exc.hresult == SCARD_W_REMOVED_CARD
    text = str(exc)
    if platform.system() == "Windows":
        expected = (
            "The smart card has been removed, "
            "so that further communication is not possible.  (0x80100069)"
        )
    else:
        expected = "Card was removed. (0x80100069)"

    assert text == expected

    exc = CardConnectionException("", SCARD_W_REMOVED_CARD)
    assert exc.hresult == SCARD_W_REMOVED_CARD
    text = str(exc)
    assert text == expected

    # now add "foo" in the message
    expected = "foo: " + expected

    exc = CardConnectionException("foo", SCARD_W_REMOVED_CARD)
    assert exc.hresult == SCARD_W_REMOVED_CARD
    text = str(exc)
    assert text == expected

    exc = CardConnectionException(hresult=SCARD_W_REMOVED_CARD, message="foo")
    assert exc.hresult == SCARD_W_REMOVED_CARD
    text = str(exc)
    assert text == expected


def test_hresult():
    hresult, hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
    if hresult == SCARD_S_SUCCESS:
        hresult, _, _ = SCardConnect(
            hcontext, "INVALID READER NAME", SCARD_SHARE_SHARED, SCARD_PROTOCOL_ANY
        )
        assert hresult == SCARD_E_UNKNOWN_READER
    else:
        assert hresult == SCARD_E_NO_SERVICE


@pytest.mark.parametrize("arg", ([0], "foo"))
def test_wrong_type(arg):
    # SCardEstablishContext() argument should be int
    with pytest.raises(TypeError):
        SCardEstablishContext(arg)


def test_card_request_timeout_exception():
    exc = CardRequestTimeoutException()
    assert str(exc) == "Time-out during card request"
    exc = CardRequestTimeoutException(SCARD_E_NOT_TRANSACTED)
    if platform.system() == "Windows":
        expected = (
            "Time-out during card request: "
            "An attempt was made to end a non-existent transaction.  (0x80100016)"
        )
    else:
        expected = "Time-out during card request: Transaction failed. (0x80100016)"
    assert str(exc) == expected
    exc = CardRequestTimeoutException(hresult=SCARD_E_NOT_TRANSACTED)
    assert str(exc) == expected


def test_invalid_atr_mask_length_exception():
    exc = InvalidATRMaskLengthException("3B 00")
    assert str(exc) == "Invalid ATR mask length: 3B 00"


def test_no_card_exception():
    exc = NoCardException("foo bar", -1)
    assert str(exc) == "foo bar"

    exc = NoCardException("foo bar", SCARD_E_NOT_TRANSACTED)
    if platform.system() == "Windows":
        expected = (
            "foo bar: "
            "An attempt was made to end a non-existent transaction.  (0x80100016)"
        )
    else:
        expected = "foo bar: Transaction failed. (0x80100016)"
    assert str(exc) == expected



"""wxPython toolbar with reader icons implementing ReaderObserver.

__author__ = "gemalto https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import wx

from smartcard.ReaderMonitoring import ReaderMonitor, ReaderObserver
from smartcard.wx import ICO_READER, ICO_SMARTCARD


class ReaderComboBox(wx.ComboBox, ReaderObserver):

    def __init__(self, parent):
        """Constructor. Registers as ReaderObserver to get
        notifications of reader insertion/removal."""
        wx.ComboBox.__init__(
            self,
            parent,
            wx.NewId(),
            size=(170, -1),
            style=wx.CB_DROPDOWN | wx.CB_SORT,
            choices=[],
        )

        # register as a ReaderObserver; we will get
        #  notified of added/removed readers
        self.readermonitor = ReaderMonitor()
        self.readermonitor.addObserver(self)

    def update(self, observable, handlers):
        """Toolbar ReaderObserver callback that is notified when
        readers are added or removed."""
        addedreaders, removedreaders = handlers
        for reader in addedreaders:
            item = self.Append(str(reader))
            self.SetClientData(item, reader)
        for reader in removedreaders:
            item = self.FindString(str(reader))
            if wx.NOT_FOUND != item:
                self.Delete(item)
        selection = self.GetSelection()
        # if wx.NOT_FOUND == selection:
        #    self.SetSelection(0)


class ReaderToolbar(wx.ToolBar):
    """ReaderToolbar. Contains controls to select a reader from a listbox
    and connect to the cards."""

    def __init__(self, parent):
        """Constructor, creating the reader toolbar."""
        wx.ToolBar.__init__(
            self,
            parent,
            pos=wx.DefaultPosition,
            size=wx.DefaultSize,
            style=wx.SIMPLE_BORDER | wx.TB_HORIZONTAL | wx.TB_FLAT | wx.TB_TEXT,
            name="Reader Toolbar",
        )

        # create bitmaps for toolbar
        tsize = (16, 16)
        if None != ICO_READER:
            bmpReader = wx.Bitmap(ICO_READER, wx.BITMAP_TYPE_ICO)
        else:
            bmpReader = wx.ArtProvider_GetBitmap(wx.ART_HELP_BOOK, wx.ART_OTHER, tsize)
        if None != ICO_SMARTCARD:
            bmpCard = wx.Bitmap(ICO_SMARTCARD, wx.BITMAP_TYPE_ICO)
        else:
            bmpCard = wx.ArtProvider_GetBitmap(wx.ART_HELP_BOOK, wx.ART_OTHER, tsize)
        self.readercombobox = ReaderComboBox(self)

        # create and add controls
        self.AddSimpleTool(
            10, bmpReader, "Select smart card reader", "Select smart card reader"
        )
        self.AddControl(self.readercombobox)
        self.AddSeparator()
        self.AddSimpleTool(20, bmpCard, "Connect to smartcard", "Connect to smart card")
        self.AddSeparator()

        self.Realize()



#! /usr/bin/env python3
"""
Sample for python PCSC wrapper module: send a Control Code to a card or
reader

__author__ = "Ludovic Rousseau"

Copyright 2007-2010 Ludovic Rousseau
Author: Ludovic Rousseau, mailto:ludovic.rousseau@free.fr

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from smartcard.scard import *
from smartcard.util import toASCIIString, toBytes, toHexString

try:
    hresult, hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
    if hresult != SCARD_S_SUCCESS:
        raise error("Failed to establish context: " + SCardGetErrorMessage(hresult))
    print("Context established!")

    try:
        hresult, readers = SCardListReaders(hcontext, [])
        if hresult != SCARD_S_SUCCESS:
            raise error("Failed to list readers: " + SCardGetErrorMessage(hresult))
        print("PCSC Readers:", readers)

        if len(readers) < 1:
            raise error("No smart card readers")

        for zreader in readers:

            print("Trying to Control reader:", zreader)

            try:
                hresult, hcard, dwActiveProtocol = SCardConnect(
                    hcontext, zreader, SCARD_SHARE_DIRECT, SCARD_PROTOCOL_T0
                )
                if hresult != SCARD_S_SUCCESS:
                    raise error("Unable to connect: " + SCardGetErrorMessage(hresult))
                print("Connected with active protocol", dwActiveProtocol)

                try:
                    if "winscard" == resourceManager:
                        # IOCTL_SMARTCARD_GET_ATTRIBUTE = SCARD_CTL_CODE(2)
                        hresult, response = SCardControl(
                            hcard,
                            SCARD_CTL_CODE(2),
                            toBytes("%.8lx" % SCARD_ATTR_VENDOR_NAME),
                        )
                        if hresult != SCARD_S_SUCCESS:
                            raise error(
                                "SCardControl failed: " + SCardGetErrorMessage(hresult)
                            )
                        print("SCARD_ATTR_VENDOR_NAME:", toASCIIString(response))
                    elif "pcsclite" == resourceManager:
                        # get feature request
                        hresult, response = SCardControl(
                            hcard, SCARD_CTL_CODE(3400), []
                        )
                        if hresult != SCARD_S_SUCCESS:
                            raise error(
                                "SCardControl failed: " + SCardGetErrorMessage(hresult)
                            )
                        print("CM_IOCTL_GET_FEATURE_REQUEST:", toHexString(response))
                finally:
                    hresult = SCardDisconnect(hcard, SCARD_UNPOWER_CARD)
                    if hresult != SCARD_S_SUCCESS:
                        raise error(
                            "Failed to disconnect: " + SCardGetErrorMessage(hresult)
                        )
                    print("Disconnected")

            except error as message:
                print(error, message)

    finally:
        hresult = SCardReleaseContext(hcontext)
        if hresult != SCARD_S_SUCCESS:
            raise error("Failed to release context: " + SCardGetErrorMessage(hresult))
        print("Released context.")

except error as e:
    print(e)

import sys

if "win32" == sys.platform:
    print("press Enter to continue")
    sys.stdin.read(1)



import platform

from smartcard.scard import (
    SCARD_F_INTERNAL_ERROR,
    SCARD_S_SUCCESS,
    SCardGetErrorMessage,
)


def test_scard_get_error_message():
    res = SCardGetErrorMessage(SCARD_S_SUCCESS)

    # do not test on Windows
    # the error messages are different and localized
    if platform.system() == "Windows":
        return

    expected = "Command successful."
    assert res == expected

    res = SCardGetErrorMessage(SCARD_F_INTERNAL_ERROR)
    expected = "Internal error."
    assert res == expected

    res = SCardGetErrorMessage(1)
    expected = "Unknown error: 0x00000001"
    # macOS bug not yet fixed
    macos_bug_expected = "Unkown error: 0x00000001"
    assert res in (expected, macos_bug_expected)
